#include <stdio.h>
#include <iostream>
#include <stdlib.h>
#include <string.h>
#include <cuda.h>
#include <time.h>
#include <sys/time.h>
//#include <cuda_runtime_api.h>
#include "Chi2.h"

int main(int argc ,char* argv[]) {

	FILE *fp;
	size_t size;
	  
	/* Initialize rows, cols, ncases, ncontrols from the user */
	unsigned int rows=atoi(argv[2]);
	unsigned int cols=atoi(argv[3]);
	int ncases=atoi(argv[4]);
	int ncontrols=atoi(argv[5]);
	int CUDA_DEVICE = atoi(argv[6]);
	int THREADS = atoi(argv[7]);
	printf("Individuals=%d SNPs=%d cases=%d controls=%d CUDA_DEVICE=%d\n",rows,cols,ncases,ncontrols,CUDA_DEVICE);

	cudaError err = cudaSetDevice(CUDA_DEVICE);
	if(err != cudaSuccess) { printf("Error setting CUDA DEVICE\n"); exit(EXIT_FAILURE); }

	/*Host variable declaration */

	//int THREADS = 32;
	int BLOCKS;
	float* host_results = (float*) malloc(cols * sizeof(float)); 
	struct timeval starttime, endtime;
	clock_t start, end;
	float seconds;
	unsigned int jobs; 
	unsigned long i;
	unsigned long ulone = 1;
	unsigned long ultwo = 2; 

	/*Kernel variable declaration */
	unsigned char *dev_dataT;
	float *results;
        char *line = NULL; size_t len = 0;
	char *token, *saveptr;

	start = clock();

	/* Validation to check if the data file is readable */
	fp = fopen(argv[1], "r");
	if (fp == NULL) {
    		printf("Cannot Open the File");
		return 0;
	}

	size = (size_t)((size_t)rows * (size_t)cols);
	printf("Size of the data = %lu\n",size);

	fflush(stdout);

	unsigned char *dataT = (unsigned char*)malloc((size_t)size);

	if(dataT == NULL) {
	        printf("ERROR: Memory for data not allocated.\n");
	}

        gettimeofday(&starttime, NULL);

        /* Transfer the SNP Data from the file to CPU Memory */
        i=0;
	while (getline(&line, &len, fp) != -1) {
                token = strtok_r(line, " ", &saveptr);
                while(token != NULL){
                        dataT[i] = *token;
                        i++;
                        token = strtok_r(NULL, " ", &saveptr);
                }
/*
                cur=0; read=-1;
                token = strtok(line, " ");
                while(sscanf(line+cur, "%d%n", &tmp, &read)==1){
                        dataT[i] = (char)(((int)'0')+tmp);
                        cur += read;
                        i++;
                }
*/
  	}
	fclose(fp);
        printf("read data\n");
        fflush(stdout);

        gettimeofday(&endtime, NULL);
        seconds+=((double)endtime.tv_sec+(double)endtime.tv_usec/1000000)-((double)starttime.tv_sec+(double)starttime.tv_usec/1000000);

        printf("time to read data = %f\n", seconds);

	/* allocate the Memory in the GPU for SNP data */	   
        gettimeofday(&starttime, NULL);
	err = cudaMalloc((unsigned char**) &dev_dataT, (size_t) size * (size_t) sizeof(unsigned char) );
	if(err != cudaSuccess) { printf("Error mallocing data on GPU device\n"); }
        gettimeofday(&endtime, NULL); seconds=((double)endtime.tv_sec+(double)endtime.tv_usec/1000000)-((double)starttime.tv_sec+(double)starttime.tv_usec/1000000);
	printf("time for cudamalloc=%f\n", seconds);

        gettimeofday(&starttime, NULL);
	err = cudaMalloc((float**) &results, cols * sizeof(float) );
	if(err != cudaSuccess) { printf("Error mallocing results on GPU device\n"); }
        gettimeofday(&endtime, NULL); seconds=((double)endtime.tv_sec+(double)endtime.tv_usec/1000000)-((double)starttime.tv_sec+(double)starttime.tv_usec/1000000);
	printf("time for cudamalloc2=%f\n", seconds);

	/*Copy the SNP data to GPU */
        gettimeofday(&starttime, NULL);
	err = cudaMemcpy(dev_dataT, dataT, (size_t)size * (size_t)sizeof(unsigned char), cudaMemcpyHostToDevice);
	if(err != cudaSuccess) { printf("Error copying data to GPU\n"); }
        gettimeofday(&endtime, NULL); seconds=((double)endtime.tv_sec+(double)endtime.tv_usec/1000000)-((double)starttime.tv_sec+(double)starttime.tv_usec/1000000);
	printf("time to copy to GPU=%f\n", seconds);

	jobs = cols;
	BLOCKS = (jobs + THREADS - 1)/THREADS;

        gettimeofday(&starttime, NULL);
	/*Calling the kernel function */
	kernel<<<BLOCKS,THREADS>>>(rows,cols,ncases,ncontrols,dev_dataT,results);
        gettimeofday(&endtime, NULL); seconds=((double)endtime.tv_sec+(double)endtime.tv_usec/1000000)-((double)starttime.tv_sec+(double)starttime.tv_usec/1000000);
	printf("time for kernel=%f\n", seconds);
		
	/*Copy the results back in host*/
	cudaMemcpy(host_results,results,cols * sizeof(float),cudaMemcpyDeviceToHost);
		
	for(int k = 0; k < jobs; k++) {
		printf("%f ", host_results[k]);
	}
	printf("\n");

	cudaFree( dev_dataT );
	cudaFree( results );

	end = clock();
	seconds = (float)(end - start) / CLOCKS_PER_SEC;
	printf("Total time = %f\n", seconds);

	return 0;

}
